/*
 * PSP Software Development Kit - http://www.pspdev.org
 * -----------------------------------------------------------------------
 * Licensed under the BSD license, see LICENSE in PSPSDK root for details.
 *
 * main.c - Fast Keyboard
 *
 * Copyright (c) 2005 Frank Buss <fb@frank-buss.de> (aka Shine)
 *
 * $Id$
 */
#include <pspkernel.h>
#include <pspdebug.h>
#include <pspdisplay.h>
#include <pspctrl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* Define the module info section */
PSP_MODULE_INFO("FASTKEYBOARD", 0, 1, 1);

/* Define the main thread's attribute value (optional) */
PSP_MAIN_THREAD_ATTR(THREAD_ATTR_USER | THREAD_ATTR_VFPU);

/* Define printf, just to make typing easier */
#define printf	pspDebugScreenPrintf

/* Exit callback */
int exit_callback(void)
{
	sceKernelExitGame();

	return 0;
}

/* Callback thread */
void CallbackThread(void *arg)
{
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", exit_callback, NULL);
	sceKernelRegisterExitCallback(cbid);

	sceKernelSleepThreadCB();
}

/* Sets up the callback thread and returns its thread id */
int SetupCallbacks(void)
{
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", CallbackThread, 0x11, 0xFA0, 0, 0);
	if(thid >= 0)
	{
		sceKernelStartThread(thid, 0, 0);
	}

	return thid;
}

#define PAD_MASK (PSP_CTRL_UP | PSP_CTRL_RIGHT | PSP_CTRL_DOWN | PSP_CTRL_LEFT)

#define ACTION_DEL -1
#define ACTION_ENTER -2
#define ACTION_SPACE -3

typedef struct
{
	int mask;
	int action;
} Key;

Key withoutStartKeys[100];
int withoutStartKeysCount = 0;

Key withStartKeys[100];
int withStartKeysCount = 0;

#define putChar(x, y, c) pspDebugScreenPutChar((x)*7, (y)*8, 0xFFFFFFFF, c)

static u32* g_vram_base = (u32*)(0x04000000+0x40000000);

void putPixel(int x, int y, int color)
{
	g_vram_base[x + y * 512] = color;
}

void drawLine(int x0, int y0, int x1, int y1, int color)
{
	#define SWAP(a, b) tmp = a; a = b; b = tmp;
	int x, y, e, dx, dy, tmp;
	if (x0 > x1) {
		SWAP(x0, x1);
		SWAP(y0, y1);
	}
	e = 0;
	x = x0;
	y = y0;
	dx = x1 - x0;
	dy = y1 - y0;
	if (dy >= 0) {
		if (dx >= dy) {
			for (x = x0; x <= x1; x++) {
				putPixel(x, y, color);
				if (2 * (e + dy) < dx) {
					e += dy;
				} else {
					y++;
					e += dy - dx;
				}
			}
		} else {
			for (y = y0; y <= y1; y++) {
				putPixel(x, y, color);
				if (2 * (e + dx) < dy) {
					e += dx;
				} else {
					x++;
					e += dx - dy;
				}
			}
		}
	} else {
		if (dx >= -dy) {
			for (x = x0; x <= x1; x++) {
				putPixel(x, y, color);
				if (2 * (e + dy) > -dx) {
					e += dy;
				} else {
					y--;
					e += dy + dx;
				}
			}
		} else {   	
			SWAP(x0, x1);
			SWAP(y0, y1);
			x = x0;
			dx = x1 - x0;
			dy = y1 - y0;
			for (y = y0; y <= y1; y++) {
				putPixel(x, y, color);
				if (2 * (e + dx) > -dy) {
					e += dx;
				} else {
					x--;
					e += dx + dy;
				}
			}
		}
	}
}

void writeByte(int fd, unsigned char data)
{
	sceIoWrite(fd, &data, 1);
}

void screenshot()
{
	const char tgaHeader[] = {0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0};
	const int width = 480;
	const int lineWidth = 512;
	const int height = 272;
	unsigned char lineBuffer[width*4];
	u32* vram = g_vram_base;
	int x, y;
	int fd = sceIoOpen("ms0:/screenshot.tga", PSP_O_CREAT | PSP_O_TRUNC | PSP_O_WRONLY, 0777);
	if (!fd) return;
	sceIoWrite(fd, tgaHeader, sizeof(tgaHeader));
	writeByte(fd, width & 0xff);
	writeByte(fd, width >> 8);
	writeByte(fd, height & 0xff);
	writeByte(fd, height >> 8);
	writeByte(fd, 24);
	writeByte(fd, 0);
	for (y = height - 1; y >= 0; y--) {
		for (x = 0; x < width; x++) {
			u32 color = vram[y * lineWidth + x];
			unsigned char red = color & 0xff;
			unsigned char green = (color >> 8) & 0xff;
			unsigned char blue = (color >> 16) & 0xff;
			lineBuffer[3*x] = blue;
			lineBuffer[3*x+1] = green;
			lineBuffer[3*x+2] = red;
		}
		sceIoWrite(fd, lineBuffer, width * 3);
	}
	sceIoClose(fd);
}

void drawRect(int x0, int y0, int x1, int y1, int color)
{
	drawLine(x0, y0, x1, y0, color);
	drawLine(x1, y0, x1, y1, color);
	drawLine(x1, y1, x0, y1, color);
	drawLine(x0, y1, x0, y0, color);
}
	
void highlightSet(int withoutStart)
{
	int x0 = 3, y0 = 67, x1 = 394, y1 = 152;
	drawRect(x0, y0, x1, y1, withoutStart ? 0xFFFFFFFF : 0);
	y0 += 12*8;
	y1 += 12*8;
	drawRect(x0, y0, x1, y1, withoutStart ? 0 : 0xFFFFFFFF);
}

void highlightKeys(int withoutStart, int mask, int color)
{
	const int w = 6;
	int x = 2;
	int y = 10;
		
	int padMask = mask & PAD_MASK;
	if (padMask == PSP_CTRL_UP) x += w;
	if (padMask == PSP_CTRL_RIGHT) x += 2*w;
	if (padMask == PSP_CTRL_DOWN) x += 3*w;
	if (padMask == PSP_CTRL_LEFT) x += 4*w;

	if (padMask == (PSP_CTRL_UP|PSP_CTRL_LEFT)) x += 5*w;
	if (padMask == (PSP_CTRL_UP|PSP_CTRL_RIGHT)) x += 6*w;
	if (padMask == (PSP_CTRL_DOWN|PSP_CTRL_LEFT)) x += 7*w;
	if (padMask == (PSP_CTRL_DOWN|PSP_CTRL_RIGHT)) x += 8*w;
	
	if (!withoutStart) y += 12;
	
	x *= 7;
	y *= 8;
	int x0 = x - 5;
	int y0 = y - 3;
	int x1 = x0 + 4*7 + 3;
	int y1 = y0 + 7*8 + 6;
	
	drawRect(x0, y0, x1, y1, color);
}	

void drawKeys(Key* keys, int count, int y)
{
	const int w = 6;
	int i;
	for (i = 0; i < count; i++) {
		int mask = keys[i].mask;
		int action = keys[i].action;
		if (action < 0) action = ' ';
		int x = 2;
		int y2 = y;
		
		int padMask = mask & PAD_MASK;
		if (padMask == PSP_CTRL_UP) x += w;
		if (padMask == PSP_CTRL_RIGHT) x += 2*w;
		if (padMask == PSP_CTRL_DOWN) x += 3*w;
		if (padMask == PSP_CTRL_LEFT) x += 4*w;

		if (padMask == (PSP_CTRL_UP|PSP_CTRL_LEFT)) x += 5*w;
		if (padMask == (PSP_CTRL_UP|PSP_CTRL_RIGHT)) x += 6*w;
		if (padMask == (PSP_CTRL_DOWN|PSP_CTRL_LEFT)) x += 7*w;
		if (padMask == (PSP_CTRL_DOWN|PSP_CTRL_RIGHT)) x += 8*w;

		if (mask & PSP_CTRL_UP) {
			int x0 = (x + 1) * 7 + 1;
			int y0 = y2 * 8 + 2;
			drawLine(x0 - 3, y0, x0, y0 - 3, 0xffffffff);
			drawLine(x0, y0 - 3, x0 + 3, y0, 0xffffffff);
			drawLine(x0 - 3, y0, x0 + 3, y0, 0xffffffff);
		}
		if (mask & PSP_CTRL_RIGHT) {
			int x0 = (x + 3) * 7 - 2;
			int y0 = (y2 + 1) * 8 + 2;
			drawLine(x0, y0 - 3, x0 + 3, y0, 0xffffffff);
			drawLine(x0 + 3, y0, x0, y0 + 3, 0xffffffff);
			drawLine(x0, y0 - 3, x0, y0 + 3, 0xffffffff);
		}
		if (mask & PSP_CTRL_DOWN) {
			int x0 = (x + 1) * 7 + 1;
			int y0 = (y2 + 2) * 8 + 2;
			drawLine(x0 - 3, y0, x0, y0 + 3, 0xffffffff);
			drawLine(x0, y0 + 3, x0 + 3, y0, 0xffffffff);
			drawLine(x0 - 3, y0, x0 + 3, y0, 0xffffffff);
		}
		if (mask & PSP_CTRL_LEFT) {
			int x0 = x * 7 + 1;
			int y0 = (y2 + 1) * 8 + 2;
			drawLine(x0, y0 - 3, x0 - 3, y0, 0xffffffff);
			drawLine(x0 - 3, y0, x0, y0 + 3, 0xffffffff);
			drawLine(x0, y0 - 3, x0, y0 + 3, 0xffffffff);
		}
		if (mask & PSP_CTRL_TRIANGLE) {
			putChar(x+1, y2+4, action);
		}
		if (mask & PSP_CTRL_CIRCLE) {
			putChar(x+2, y2+5, action);
		}
		if (mask & PSP_CTRL_CROSS) {
			putChar(x+1, y2+6, action);
		}
		if (mask & PSP_CTRL_SQUARE) {
			putChar(x, y2+5, action);
		}
	}
}

int main(void)
{
	SceCtrlData pad;
	u32 buttonsold = 0;

	SetupCallbacks();
	pspDebugScreenInit();

	sceCtrlSetSamplingCycle(0);
	sceCtrlSetSamplingMode(PSP_CTRL_MODE_DIGITAL);

	printf ("Fast Keyboard implementation by Shine\n\n");

	FILE* file = fopen("ms0:/layout.txt", "r");
	if (!file) {
		printf("ms0:/layout.txt not found\n");
		while (1) sceDisplayWaitVblankStart(); 
	}
	int withoutStart = 1;
	while (1) {
		char line[100];
		if (!fgets(line, 100, file)) break;
		if (strlen(line) == 0) continue;
		if (line[0] == 's') {
			withoutStart = 0;
		} else {
			int mask = 0;
			char* l = line;
			while (1) {
				char c = *l;
				switch (c) {
					case '1':
						mask |= PSP_CTRL_LTRIGGER;
						break;
					case '2':
						mask |= PSP_CTRL_RTRIGGER;
						break;
					case 'u':
						mask |= PSP_CTRL_UP;
						break;
					case 'r':
						mask |= PSP_CTRL_RIGHT;
						break;
					case 'd':
						mask |= PSP_CTRL_DOWN;
						break;
					case 'l':
						mask |= PSP_CTRL_LEFT;
						break;
					case 't':
						mask |= PSP_CTRL_TRIANGLE;
						break;
					case 'o':
						mask |= PSP_CTRL_CIRCLE;
						break;
					case 'x':
						mask |= PSP_CTRL_CROSS;
						break;
					case 'q':
						mask |= PSP_CTRL_SQUARE;
						break;
				}
				l++;
				if (c == '=') break;
			}
			int action;
			if (strncmp(l, "del", 3) == 0) {
				action = ACTION_DEL;
			} else if (strncmp(l, "enter", 5) == 0) {
				action = ACTION_ENTER;
			} else if (strncmp(l, "space", 5) == 0) {
				action = ACTION_SPACE;
			} else {
				action = *l;
			}
			if (withoutStart) {
				withoutStartKeys[withoutStartKeysCount].mask = mask;
				withoutStartKeys[withoutStartKeysCount].action = action;
				withoutStartKeysCount++;
			} else {
				withStartKeys[withStartKeysCount].mask = mask;
				withStartKeys[withStartKeysCount].action = action;
				withStartKeysCount++;
			}
		}
	}
	fclose(file);
	
	drawKeys(withoutStartKeys, withoutStartKeysCount, 10);
	drawKeys(withStartKeys, withStartKeysCount, 22);
	
	char buf[2];
	buf[1] = 0;
	withoutStart = 1;
	highlightKeys(1, 0, 0xffffffff);
	while (1) {
		highlightSet(withoutStart);
		sceCtrlReadBufferPositive(&pad, 1);
		if (pad.Buttons != buttonsold) {
			if (pad.Buttons & PSP_CTRL_START) {
				highlightKeys(withoutStart, buttonsold & PAD_MASK, 0);
				withoutStart = !withoutStart;
				buttonsold = pad.Buttons;
			} else {
				buttonsold = pad.Buttons;
				Key* keys = withoutStart ? withoutStartKeys : withStartKeys;
				int c = withoutStart ? withoutStartKeysCount : withStartKeysCount;
				int i;
				for (i = 0; i < c; i++) {
					int mask = keys[i].mask;
					highlightKeys(withoutStart, mask, (mask & PAD_MASK) == (buttonsold & PAD_MASK) ? 0xffffffff : 0);
					if (mask == buttonsold) {
						int action = keys[i].action;
						switch (action) {
							case ACTION_DEL:
								// TODO
								break;
							case ACTION_ENTER:
								// TODO
								break;
							case ACTION_SPACE:
								printf(" ");
								break;
							default:
								buf[0] = action;
								printf("%s", buf);
								break;
						}
					}
				}
			}
		}
		sceDisplayWaitVblankStart(); 
	}

	return 0;
}
